#
# Test for the util.lvm2 module
#

import io
import os

import pytest

from osbuild.util import pe32p

EFI_STUB = "/usr/lib/systemd/boot/efi/linuxx64.efi.stub"


def have_efi_stub() -> bool:
    return os.path.exists(EFI_STUB)


@pytest.mark.skipif(not have_efi_stub(), reason="require systemd efi stub")
def test_basic():
    with open(EFI_STUB, "rb") as f:
        coff = pe32p.read_coff_header(f)
        assert coff
        opt = pe32p.read_optional_header(f, coff)
        assert opt
        sections = pe32p.read_sections(f, coff)

    assert sections, "No sections found in stub"


@pytest.mark.skipif(not have_efi_stub(), reason="require systemd efi stub")
def test_basic_no_coff():
    # check the API versions that re-reads the CoffHeader
    with open(EFI_STUB, "rb") as f:
        coff = pe32p.read_coff_header(f)
        assert coff
        f.seek(0, io.SEEK_SET)
        opt = pe32p.read_optional_header(f)
        f.seek(0, io.SEEK_SET)
        assert opt
        f.seek(0, io.SEEK_SET)
        sections = pe32p.read_sections(f)

    assert sections, "No sections found in stub"

    with open(EFI_STUB, "rb") as f:
        coff_check = pe32p.read_coff_header(f)
        assert coff_check
        assert coff == coff_check
        opt_check = pe32p.read_optional_header(f, coff)
        assert opt_check
        assert opt == opt_check
        sections_check = pe32p.read_sections(f, coff)
        assert sections_check
        assert sections == sections_check



#
# Tests for the 'osbuild.util.path' module
#
import os
import time
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from osbuild.util import path


@pytest.fixture(name="tempdir")
def tempdir_fixture():
    with TemporaryDirectory(prefix="path-") as tmp:
        yield tmp


def test_clamp_mtime(tempdir):
    start = int(time.time())

    tree = Path(tempdir, "tree")
    tree.mkdir()
    os.makedirs(os.path.join(tree, "a", "bunch", "of", "directories"))

    file = Path(tree, "file")
    file.touch()

    folder = Path(tree, "folder")
    folder.mkdir()

    link = Path(tree, "link")
    link.symlink_to(folder, target_is_directory=True)

    timepoint = 1644758228
    path.clamp_mtime(tree, start, timepoint)

    def ensure_mtime(target, dfd):
        stat = os.stat(target, dir_fd=dfd, follow_symlinks=False)
        assert stat.st_mtime <= timepoint

    for _, dirs, files, dfd in os.fwalk(tree):
        ensure_mtime(".", dfd)

        for d in dirs:
            ensure_mtime(d, dfd)

        for f in files:
            ensure_mtime(f, dfd)


def test_in_tree():
    cases = {
        ("/tmp/file", "/tmp", False): True,  # Simple, non-existent
        ("/etc", "/", True): True,  # Simple, should exist
        ("/tmp/../../file", "/tmp", False): False,  # Relative path escape
        ("/very/fake/directory/and/file", "/very", False): True,  # non-existent, OK
        ("/very/fake/directory/and/file", "/very", True): False,  # non-existent, not-OK
        ("../..", os.path.abspath("."), False): False,  # Relative path escape from cwd
        (".", os.path.abspath("../.."), False): True,  # cwd inside parent of parent
    }

    for args, expected in cases.items():
        assert path.in_tree(*args) == expected, args



#
# Tests for the 'osbuild.util.fscache' module.
#

# pylint: disable=protected-access

import json
import os
import tempfile

import pytest

from osbuild.util import fscache


@pytest.fixture(name="tmpdir")
def tmpdir_fixture():
    with tempfile.TemporaryDirectory(dir="/var/tmp") as tmp:
        yield tmp


def test_calculate_size(tmpdir):
    #
    # Test the `_calculate_size()` helper and verify it only includes file
    # content in its calculation.
    #

    os.mkdir(os.path.join(tmpdir, "dir"))

    assert fscache.FsCache._calculate_size(os.path.join(tmpdir, "dir")) == 0

    with open(os.path.join(tmpdir, "dir", "file"), "x", encoding="utf8") as f:
        pass

    assert fscache.FsCache._calculate_size(os.path.join(tmpdir, "dir")) == 0

    with open(os.path.join(tmpdir, "dir", "file"), "w", encoding="utf8") as f:
        f.write("foobar")

    assert fscache.FsCache._calculate_size(os.path.join(tmpdir, "dir")) == 6


def test_pathlike(tmpdir):
    #
    # Verify behavior of `__fspath__()`.
    #

    class Wrapper:
        def __init__(self, path: str):
            self._path = path

        def __fspath__(self) -> str:
            return self._path

    # Test with a plain string as argument
    dir_str: str = os.fspath(tmpdir)
    cache1 = fscache.FsCache("osbuild-test-appid", dir_str)
    assert os.fspath(cache1) == tmpdir
    assert os.path.join(cache1, "foobar") == os.path.join(tmpdir,  "foobar")

    # Test with a wrapper-type as argument
    dir_pathlike: Wrapper = Wrapper(os.fspath(tmpdir))
    cache2 = fscache.FsCache("osbuild-test-appid", dir_pathlike)
    assert os.fspath(cache2) == tmpdir
    assert os.path.join(cache2, "foobar") == os.path.join(tmpdir,  "foobar")


def test_path(tmpdir):
    #
    # Verify behavior of `_path()`.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        assert cache._path() == cache._path_cache
        assert cache._path("dir") == os.path.join(cache._path_cache, "dir")
        assert cache._path("dir", "file") == os.path.join(cache._path_cache, "dir", "file")


def test_atomic_open(tmpdir):
    #
    # Verify the `_atomic_open()` helper correctly opens existing files and
    # takes a lock.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        # Must never create files.
        with pytest.raises(OSError):
            with cache._atomic_open("file", write=False, wait=False) as f:
                pass

        # Create the file with "foo" as content.
        with open(os.path.join(tmpdir, "file"), "x", encoding="utf8") as f:
            f.write("foo")

        # Open and acquire a write-lock. Then verify a read-lock fails.
        with cache._atomic_open("file", write=True, wait=False):
            with pytest.raises(BlockingIOError):
                with cache._atomic_open("file", write=False, wait=False):
                    pass


def test_atomic_file(tmpdir):
    #
    # Verify behavior of `_atomic_file()` as replacement for `O_TMPFILE`.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        rpath_store = cache._dirname_objects
        path_store = os.path.join(cache._path_cache, rpath_store)

        # Initially the store is empty.
        assert len(list(os.scandir(path_store))) == 0

        # Create a file and verify there is almost exactly 1 file in the store.
        with cache._atomic_file(os.path.join(rpath_store, "file"), rpath_store) as f:
            assert len(list(os.scandir(path_store))) == 1
            f.write("foo")
        assert len(list(os.scandir(path_store))) == 1

        # Verify `ignore_exist=False` works as expected.
        with pytest.raises(OSError):
            with cache._atomic_file(os.path.join(rpath_store, "file"), rpath_store) as f:
                # Temporarily, there will be 2 files.
                assert len(list(os.scandir(path_store))) == 2
                f.write("bar")
        assert len(list(os.scandir(path_store))) == 1
        with open(os.path.join(path_store, "file"), "r", encoding="utf8") as f:
            assert f.read() == "foo"

        # Verify `ignore_exist=True` works as expected.
        with cache._atomic_file(os.path.join(rpath_store, "file"), rpath_store, ignore_exist=True) as f:
            f.write("bar")
        assert len(list(os.scandir(path_store))) == 1
        with open(os.path.join(path_store, "file"), "r", encoding="utf8") as f:
            assert f.read() == "foo"

        # Verify `replace=True`.
        with cache._atomic_file(os.path.join(rpath_store, "file"), rpath_store, replace=True) as f:
            f.write("bar")
        assert len(list(os.scandir(path_store))) == 1
        with open(os.path.join(path_store, "file"), "r", encoding="utf8") as f:
            assert f.read() == "bar"

        # Combining `replace` and `ignore_exist` is not allowed.
        with pytest.raises(AssertionError):
            with cache._atomic_file(
                    os.path.join(rpath_store, "file"),
                    rpath_store,
                    replace=True,
                    ignore_exist=True,
            ) as f:
                pass


def test_atomic_dir(tmpdir):
    #
    # Verify the `_atomic_dir()` helper correctly creates anonymous files
    # and yields the name and lock-file.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        # The relative-path must exist, so expect an error if it does not.
        with pytest.raises(OSError):
            cache._atomic_dir("dir")

        assert len(list(os.scandir(os.path.join(tmpdir, cache._dirname_objects)))) == 0

        (name, lockfd) = cache._atomic_dir(cache._dirname_objects)
        assert name.startswith("uuid-")
        assert len(name) == 37
        assert lockfd >= 0
        os.close(lockfd)

        assert len(list(os.scandir(os.path.join(tmpdir, cache._dirname_objects)))) == 1


def test_scaffolding(tmpdir):
    #
    # Verify that the cache creates scaffolding when entered.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    assert len(list(os.scandir(tmpdir))) == 0

    with cache:
        pass

    assert len(list(os.scandir(tmpdir))) == 6
    assert len(list(os.scandir(os.path.join(tmpdir, cache._dirname_objects)))) == 0
    assert len(list(os.scandir(os.path.join(tmpdir, cache._dirname_stage)))) == 0

    with open(os.path.join(tmpdir, cache._filename_cache_tag), "r", encoding="utf8") as f:
        assert len(f.read()) > 0
    with open(os.path.join(tmpdir, cache._filename_cache_info), "r", encoding="utf8") as f:
        assert json.load(f) == {"version": 1}
    with open(os.path.join(tmpdir, cache._filename_cache_lock), "r", encoding="utf8") as f:
        assert f.read() == ""
    with open(os.path.join(tmpdir, cache._filename_cache_size), "r", encoding="utf8") as f:
        assert f.read() == "0"


def test_cachedir_tag(tmpdir):
    #
    # Verify compatibility to the cachedir-tag specification.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    with cache:
        pass

    with open(os.path.join(tmpdir, "CACHEDIR.TAG"), "r", encoding="utf8") as f:
        assert f.read(43) == "Signature: 8a477f597d28d172789f06886806bc55"


def test_cache_info(tmpdir):
    #
    # Verify that the cache reads and augments cache information. Also verify
    # the default values.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    with cache:
        assert cache._info == fscache.FsCacheInfo(version=1)
        assert cache.info == cache._info

        assert cache.info.maximum_size is None
        assert cache.info.creation_boot_id is None
        cache.info = fscache.FsCacheInfo(maximum_size=1024)
        assert cache.info.maximum_size == 1024
        assert cache.info.creation_boot_id is None
        cache.info = fscache.FsCacheInfo(creation_boot_id="0"*32)
        assert cache.info.maximum_size == 1024
        assert cache.info.creation_boot_id == "0"*32
        cache.info = fscache.FsCacheInfo(maximum_size=2048, creation_boot_id="1"*32)
        assert cache.info.maximum_size == 2048
        assert cache.info.creation_boot_id == "1"*32

    assert not fscache.FsCacheInfo().to_json()
    assert fscache.FsCacheInfo(creation_boot_id="0"*32).to_json() == {
        "creation-boot-id": "0"*32,
    }
    assert fscache.FsCacheInfo(creation_boot_id="0"*32, maximum_size=1024).to_json() == {
        "creation-boot-id": "0"*32,
        "maximum-size": 1024,
    }

    assert fscache.FsCacheInfo.from_json({}) == fscache.FsCacheInfo()
    assert fscache.FsCacheInfo.from_json(None) == fscache.FsCacheInfo()
    assert fscache.FsCacheInfo.from_json("foobar") == fscache.FsCacheInfo()
    assert fscache.FsCacheInfo.from_json({
        "creation-boot-id": "0"*32,
    }) == fscache.FsCacheInfo(creation_boot_id="0"*32)
    assert fscache.FsCacheInfo.from_json({
        "creation-boot-id": "0"*32,
        "maximum-size": 1024,
    }) == fscache.FsCacheInfo(creation_boot_id="0"*32, maximum_size=1024)
    assert fscache.FsCacheInfo.from_json({
        "creation-boot-id": "0"*32,
        "maximum-size": 1024,
    }) == fscache.FsCacheInfo(creation_boot_id="0"*32, maximum_size=1024)
    assert fscache.FsCacheInfo.from_json({
        "creation-boot-id": "0"*32,
        "unknown0": "foobar",
        "unknown1": ["foo", "bar"],
    }) == fscache.FsCacheInfo(creation_boot_id="0"*32)


def test_store(tmpdir):
    #
    # API tests for the `store()` method.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    with pytest.raises(AssertionError):
        with cache.store("foobar"):
            pass

    with cache:
        with pytest.raises(ValueError):
            with cache.store(""):
                pass


def test_load(tmpdir):
    #
    # API tests for the `load()` method.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    with pytest.raises(AssertionError):
        with cache.load("foobar"):
            pass

    with cache:
        with pytest.raises(ValueError):
            with cache.load(""):
                pass


def test_store_tree(tmpdir):
    #
    # API tests for the `store_tree()` method.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)

    with pytest.raises(AssertionError):
        cache.store_tree("foobar", "invalid/dir")

    with cache:
        cache.info = cache.info._replace(maximum_size=1024*1024*1024)

        with pytest.raises(ValueError):
            cache.store_tree("", "invalid/dir")
        with pytest.raises(RuntimeError):
            cache.store_tree("key", "invalid/dir")

        with tempfile.TemporaryDirectory(dir="/var/tmp") as tmp:
            with open(os.path.join(tmp, "outside"), "x", encoding="utf8") as f:
                f.write("foo")
            os.mkdir(os.path.join(tmp, "tree"))
            with open(os.path.join(tmp, "tree", "inside"), "x", encoding="utf8") as f:
                f.write("bar")
            with open(os.path.join(tmp, "tree", "more-inside"), "x", encoding="utf8") as f:
                f.write("foobar")

            cache.store_tree("key", os.path.join(tmp, "tree"))

            with cache.load("key") as rpath:
                assert len(list(os.scandir(os.path.join(cache, rpath)))) == 1
                assert len(list(os.scandir(os.path.join(cache, rpath, "tree")))) == 2
                with open(os.path.join(cache, rpath, "tree", "inside"), "r", encoding="utf8") as f:
                    assert f.read() == "bar"
                with open(os.path.join(cache, rpath, "tree", "more-inside"), "r", encoding="utf8") as f:
                    assert f.read() == "foobar"


def test_basic(tmpdir):
    #
    # A basic cache store+load test.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        cache.info = cache.info._replace(maximum_size=1024)

        with cache.stage() as rpath:
            with open(os.path.join(tmpdir, rpath, "bar"), "x", encoding="utf8") as f:
                f.write("foobar")

        with pytest.raises(fscache.FsCache.MissError):
            with cache.load("foo") as rpath:
                pass

        with cache.store("foo") as rpath:
            with open(os.path.join(tmpdir, rpath, "bar"), "x", encoding="utf8") as f:
                f.write("foobar")

        with cache.load("foo") as rpath:
            with open(os.path.join(tmpdir, rpath, "bar"), "r", encoding="utf8") as f:
                assert f.read() == "foobar"


def test_size_discard(tmpdir):
    #
    # Verify that a cache with no maximum-size configured can never store any
    # entries, but discards them immediately.
    #

    cache = fscache.FsCache("osbuild-test-appid", tmpdir)
    with cache:
        with cache.store("foo") as rpath:
            with open(os.path.join(tmpdir, rpath, "bar"), "x", encoding="utf8") as f:
                f.write("foobar")

        with pytest.raises(fscache.FsCache.MissError):
            with cache.load("foo") as rpath:
                pass



"""Helpers related to parsing"""

import re
from typing import Union


def parse_size(s: str) -> Union[int, str]:
    """Parse a size string into a number or 'unlimited'.

    Supported suffixes: kB, kiB, MB, MiB, GB, GiB, TB, TiB
    """
    units = [
        (r'^\s*(\d+)\s*kB$', 1000, 1),
        (r'^\s*(\d+)\s*KiB$', 1024, 1),
        (r'^\s*(\d+)\s*MB$', 1000, 2),
        (r'^\s*(\d+)\s*MiB$', 1024, 2),
        (r'^\s*(\d+)\s*GB$', 1000, 3),
        (r'^\s*(\d+)\s*GiB$', 1024, 3),
        (r'^\s*(\d+)\s*TB$', 1000, 4),
        (r'^\s*(\d+)\s*TiB$', 1024, 4),
        (r'^\s*(\d+)$', 1, 1),
        (r'^unlimited$', "unlimited", 1),
    ]

    for pat, base, power in units:
        m = re.fullmatch(pat, s)
        if m:
            if isinstance(base, int):
                return int(m.group(1)) * base ** power
            if base == "unlimited":
                return "unlimited"

    raise TypeError(f"invalid size value: '{s}'")



#
# Coherency tests for the 'osbuild.util.fscache' module.
#

# pylint: disable=protected-access

import contextlib
import errno
import os
import socket
import subprocess
import tempfile

import pytest

from osbuild.util import fscache, linux

from .. import test


def nfsd_available():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.settimeout(1)
        return s.connect_ex(("localhost", 2049)) == 0


@contextlib.contextmanager
def mount_nfs(src: str, dst: str):
    r = subprocess.run(
        [
            "mount",
            "-t", "nfs",
            "-o", "nosharecache,vers=4",
            src,
            dst,
        ],
        check=False,
        encoding="utf-8",
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )
    if r.returncode != 0:
        code = r.returncode
        msg = r.stdout.strip()
        raise RuntimeError(f"{msg} (code: {code})")

    try:
        yield dst
    finally:
        subprocess.run(
            ["umount", dst],
            check=True,
        )


@pytest.fixture(name="nfsmnt")
def nfsmnt_fixture():
    tmpmnt = None
    try:
        tmpmnt = tempfile.mkdtemp(prefix="osbuild-test-", dir="/var/tmp")
        with mount_nfs("localhost:/", tmpmnt) as tmpnfs:
            with tempfile.TemporaryDirectory(dir=tmpnfs) as tmpdir:
                yield tmpdir
    finally:
        os.rmdir(tmpmnt)


@pytest.fixture(name="nfsmnts")
def nfsmnts_fixture():
    tmpmnt = None
    try:
        tmpmnt = tempfile.mkdtemp(prefix="osbuild-test-", dir="/var/tmp")
        os.mkdir(os.path.join(tmpmnt, "a"))
        os.mkdir(os.path.join(tmpmnt, "b"))
        with mount_nfs("localhost:/", os.path.join(tmpmnt, "a")), \
                mount_nfs("localhost:/", os.path.join(tmpmnt, "b")):
            with tempfile.TemporaryDirectory(dir=os.path.join(tmpmnt, "a")) as tmpdir:
                dirname = os.path.basename(os.path.normpath(tmpdir))
                a = os.path.join(tmpmnt, "a", dirname)
                b = os.path.join(tmpmnt, "b", dirname)
                yield (a, b)
    finally:
        os.rmdir(os.path.join(tmpmnt, "b"))
        os.rmdir(os.path.join(tmpmnt, "a"))
        os.rmdir(tmpmnt)


@pytest.mark.skipif(not nfsd_available(), reason="NFSv4 daemon required")
@pytest.mark.skipif(not test.TestBase.can_bind_mount(), reason="Bind-mounting required")
def test_nfs_characteristics(nfsmnts):
    #
    # Test NFS Characteristic
    #
    # This mounts a single NFS instance with `nosharedcache` twice. It then
    # runs a series of custom tests to verify cache-behavior of NFS and how
    # different operations will cause stale caches and thus validate our
    # assumptions on how to avoid them.
    #

    a = nfsmnts[0]
    b = nfsmnts[1]

    # `stat` does not invalidate caches
    #
    # Write a fresh file on A, then stat it on B. This will properly re-fetch
    # all information since caches on both sides are empty. Then re-write the
    # file on A and, again, stat it on B and verify that the metadata was *NOT*
    # re-fetched, since NFS caches this information.
    # As last step read the file on B and verify that `open()` will properly
    # re-fetch all information.

    with open(os.path.join(a, "0_foo"), "x", encoding="utf8") as f:
        f.write("foo")

    assert os.stat(os.path.join(b, "0_foo")).st_size == 3

    with open(os.path.join(a, "0_foo"), "w", encoding="utf8") as f:
        f.write("foobar")

    assert os.stat(os.path.join(b, "0_foo")).st_size == 3

    with open(os.path.join(b, "0_foo"), "r", encoding="utf8") as f:
        assert f.read() == "foobar"

    assert os.stat(os.path.join(b, "0_foo")).st_size == 6

    # Lock-acquisition invalidates caches
    #
    # Create a file on A and commit it to disk. Open it for reading on A, but
    # then delete it on B. Verify on B it is gone. On A continue reading the
    # file. Stat it on A to verify the caches have not been invalidated. Then
    # acquire a read-lock on the open file and try the same again, this time
    # the cache should reflect the unlink.

    with open(os.path.join(a, "1_foo"), "x", encoding="utf8") as f:
        f.write("foo")

    with open(os.path.join(a, "1_foo"), "r", encoding="utf8") as f:
        os.unlink(os.path.join(b, "1_foo"))
        assert not os.access(os.path.join(b, "1_foo"), os.R_OK)

        assert f.read() == "foo"
        assert os.stat(f.fileno()).st_nlink == 1
        assert os.stat(os.path.join(a, "1_foo")).st_ino != 0
        assert os.access(os.path.join(a, "1_foo"), os.R_OK)

        linux.fcntl_flock(f.fileno(), linux.fcntl.F_RDLCK, wait=True)

        # The first STAT after an unlink returns a link-count of 0,
        # while every following STAT raises ESTALE. Lets try
        # verifying that, but do not depend on it and allow the
        # first STAT to raise ESTALE as well.
        try:
            assert os.stat(f.fileno()).st_nlink == 0
        except OSError as e:
            assert e.errno == errno.ESTALE

        with pytest.raises(OSError):
            os.stat(f.fileno())
        with pytest.raises(OSError):
            os.stat(os.path.join(a, "1_foo"))

        assert not os.access(os.path.join(a, "1_foo"), os.R_OK)

    # Inode changes on replacement
    #
    # Create a file, STAT it on A and B and verify they match and the
    # caches are active. Then replace the file on A and STAT again. The
    # replacement will be visible on A, but the caches on B still yield
    # the same old value. However, after opening on B, the file content
    # will yield the updated data and so will STAT.

    with open(os.path.join(a, "2_foo"), "x", encoding="utf8") as f:
        f.write("foo")

    st_a0 = os.stat(os.path.join(a, "2_foo"))
    st_b0 = os.stat(os.path.join(b, "2_foo"))
    assert st_a0.st_ino == st_b0.st_ino

    with open(os.path.join(a, "2_bar"), "x", encoding="utf8") as f:
        f.write("bar")
    os.rename(os.path.join(a, "2_bar"), os.path.join(a, "2_foo"))

    st_a1 = os.stat(os.path.join(a, "2_foo"))
    st_b1 = os.stat(os.path.join(b, "2_foo"))

    assert st_a1.st_ino != st_b1.st_ino
    assert st_b0.st_ino == st_b1.st_ino

    with open(os.path.join(b, "2_foo"), "r", encoding="utf8") as f:
        assert f.read() == "bar"

    st_a2 = os.stat(os.path.join(a, "2_foo"))
    st_b2 = os.stat(os.path.join(b, "2_foo"))

    assert st_a2.st_ino == st_b2.st_ino


def _test_atomics_with(a: str, b: str):
    with fscache.FsCache("osbuild-test-appid", a) as cache:
        cache.info = cache.info._replace(maximum_size=1024)

        # Test _atomic_open() with open+lock race
        #
        # Create a file `0_foo` and OPEN+LOCK it with _atomic_open(). Use a
        # tracer to hook between OPEN and LOCK. First time unlinke the target
        # file and recreate it. The second time, replace it instead.
        #
        # Verify that `_atomic_open()` needs 3 attempts to OPEN+LOCK the file,
        # and verify the content is ultimately correct.

        def _trace_lock(state: dict):
            # Use `open(..., "x")` to force an invalidation of NFS caches.
            # Otherwise, VFS would just deny our operations based on outdated
            # NFS-caches. This would be coherent, but overly restrictive.
            with pytest.raises(OSError):
                with open(os.path.join(b, "0_foo"), "x", encoding="utf8") as f:
                    pass

            if state["lock"] == 0:
                with open(os.path.join(b, "0_foo"), "r", encoding="utf8") as f:
                    assert f.read() == "foo"
                os.unlink(os.path.join(b, "0_foo"))
                with open(os.path.join(b, "0_foo"), "x", encoding="utf8") as f:
                    f.write("bar")
            elif state["lock"] == 1:
                with open(os.path.join(b, "0_foo"), "r", encoding="utf8") as f:
                    assert f.read() == "bar"
                with open(os.path.join(b, "0_foo2"), "x", encoding="utf8") as f:
                    f.write("foobar")
                os.rename(os.path.join(b, "0_foo2"), os.path.join(b, "0_foo"))

            state["lock"] = state["lock"] + 1

        state = {"lock": 0}
        cache._tracers = {
            "_atomic_open:lock": lambda: _trace_lock(state)
        }
        with open(os.path.join(a, "0_foo"), "x", encoding="utf8") as f:
            f.write("foo")
        with cache._atomic_open("0_foo", wait=True, write=False) as fd:
            with os.fdopen(fd, "r", closefd=False, encoding="utf8") as f:
                assert f.read() == "foobar"

        assert state["lock"] == 3


@pytest.mark.skipif(not test.TestBase.can_bind_mount(), reason="Bind-mounting required")
def test_atomics():
    #
    # Test FsCache Atomics (native)
    #
    # Verify the behavior of the `_atomic_*()` helpers of FsCache. Use the
    # trace-hooks of FsCache to trigger the race-conditions we want to test.
    #

    with tempfile.TemporaryDirectory(dir="/var/tmp") as tmpdir:
        _test_atomics_with(tmpdir, tmpdir)


@pytest.mark.skipif(not nfsd_available(), reason="NFSv4 daemon required")
@pytest.mark.skipif(not test.TestBase.can_bind_mount(), reason="Bind-mounting required")
def test_atomics_nfs(nfsmnts):
    #
    # Test FsCache Atomics (NFS)
    #
    # Same as `test_atomics()` but on NFS.
    #

    with tempfile.TemporaryDirectory(dir=nfsmnts[0]) as tmpdir:
        _test_atomics_with(tmpdir, tmpdir)

    # Preferably, we would now run the same tests on two distinct
    # NFS mounts with no shared caches. Unfortunately, this keeps
    # triggering kernel-oopses, so we disable the tests for now:

    #_test_atomics_with(nfsmnts[0], nfsmnts[1])



#
# Test for monitoring classes and integration
#

import io
import os
import sys
import tempfile
import unittest
from collections import defaultdict

import osbuild
import osbuild.meta
from osbuild.monitor import LogMonitor
from osbuild.objectstore import ObjectStore
from osbuild.pipeline import Runner

from .. import test


class TapeMonitor(osbuild.monitor.BaseMonitor):
    """Record the usage of all called functions"""

    def __init__(self):
        super().__init__(sys.stderr.fileno())
        self.counter = defaultdict(int)
        self.stages = set()
        self.asm = None
        self.results = set()
        self.logger = io.StringIO()
        self.output = None

    def begin(self, pipeline: osbuild.Pipeline):
        self.counter["begin"] += 1

    def finish(self, result):
        self.counter["finish"] += 1
        self.output = self.logger.getvalue()

    def stage(self, stage: osbuild.Stage):
        self.counter["stages"] += 1
        self.stages.add(stage.id)

    def result(self, result: osbuild.pipeline.BuildResult):
        self.counter["result"] += 1
        self.results.add(result.id)

    def log(self, message: str):
        self.counter["log"] += 1
        self.logger.write(message)


class TestMonitor(unittest.TestCase):
    @unittest.skipUnless(test.TestBase.can_bind_mount(), "root-only")
    def test_log_monitor_vfuncs(self):
        # Checks the basic functioning of the LogMonitor
        index = osbuild.meta.Index(os.curdir)

        runner = Runner(index.detect_host_runner())
        pipeline = osbuild.Pipeline("pipeline", runner=runner)
        info = index.get_module_info("Stage", "org.osbuild.noop")
        pipeline.add_stage(info, {
            "isthisthereallife": False
        })

        with tempfile.TemporaryDirectory() as tmpdir:
            storedir = os.path.join(tmpdir, "store")

            logfile = os.path.join(tmpdir, "log.txt")

            with open(logfile, "w", encoding="utf8") as log, ObjectStore(storedir) as store:
                monitor = LogMonitor(log.fileno())
                res = pipeline.run(store,
                                   monitor,
                                   libdir=os.path.abspath(os.curdir))

                with open(logfile, encoding="utf8") as f:
                    log = f.read()

            assert res
            self.assertIn(pipeline.stages[0].id, log)
            self.assertIn("isthisthereallife", log)

    @unittest.skipUnless(test.TestBase.can_bind_mount(), "root-only")
    def test_monitor_integration(self):
        # Checks the monitoring API is called properly from the pipeline
        index = osbuild.meta.Index(os.curdir)
        runner = Runner(index.detect_host_runner())

        pipeline = osbuild.Pipeline("pipeline", runner=runner)
        noop_info = index.get_module_info("Stage", "org.osbuild.noop")
        pipeline.add_stage(noop_info, {
            "isthisthereallife": False
        })
        pipeline.add_stage(noop_info, {
            "isthisjustfantasy": True
        })

        with tempfile.TemporaryDirectory() as tmpdir:
            storedir = os.path.join(tmpdir, "store")

            tape = TapeMonitor()
            with ObjectStore(storedir) as store:
                res = pipeline.run(store,
                                   tape,
                                   libdir=os.path.abspath(os.curdir))

        assert res
        self.assertEqual(tape.counter["begin"], 1)
        self.assertEqual(tape.counter["finish"], 1)
        self.assertEqual(tape.counter["stages"], 2)
        self.assertEqual(tape.counter["stages"], 2)
        self.assertEqual(tape.counter["result"], 2)
        self.assertIn(pipeline.stages[0].id, tape.stages)
        self.assertIn("isthisthereallife", tape.output)
        self.assertIn("isthisjustfantasy", tape.output)



#!/usr/bin/python3

#
# Runtime Tests for Device Host Services
#

import errno
import json
import os
import subprocess
import sys
import tempfile
from contextlib import contextmanager

import pytest

from osbuild import devices, host, meta, mounts

from ..test import TestBase


@pytest.fixture(name="tmpdir")
def tmpdir_fixture():
    with tempfile.TemporaryDirectory(prefix="test-devices-") as tmp:
        yield tmp


@contextmanager
def make_arguments(opts):
    os.mkdir("/run/osbuild/api")
    with open("/run/osbuild/api/arguments", "w", encoding="utf-8") as f:
        json.dump(opts, f)
    try:
        yield
    finally:
        os.remove("/run/osbuild/api/arguments")
        os.rmdir("/run/osbuild/api")


@contextmanager
def make_dev_tmpfs(tmpdir):
    dev_path = os.path.join(tmpdir, "dev")
    os.makedirs(dev_path)
    subprocess.run(["mount", "-t", "tmpfs", "-o", "nosuid", "none", dev_path], check=True)
    yield dev_path
    subprocess.run(["umount", "--lazy", dev_path], check=True)


def create_image(tmpdir):
    # create a file to contain an image
    tree = os.path.join(tmpdir, "tree")
    os.makedirs(tree)
    size = 2 * 1024 * 1024
    file = os.path.join(tree, "image.img")
    with open(file, "wb") as f:
        f.truncate(size)

    # Create an FS in the image
    mkfsopts = {
        "devices": {
            "device": {
                "path": file
            }
        },
        "options": {
            "label": "TEST",
            "volid": "7B7795E7",
            "fat-size": 32
        }
    }

    with make_arguments(mkfsopts):
        subprocess.run(
            [os.path.join(os.curdir, "stages",  "org.osbuild.mkfs.fat")],
            check=True,
            stdout=sys.stdout,
            stderr=sys.stderr)
    return tree, size


def mount(mgr, devpath, tree, size, mountpoint, options):
    index = meta.Index(os.curdir)
    # Device manager to open the loopback device
    devmgr = devices.DeviceManager(mgr, devpath, tree)
    # get a Device for the loopback
    dev = devices.Device(
        "loop",
        index.get_module_info(
            "Device",
            "org.osbuild.loopback"
        ),
        None,
        {
            "filename": "image.img",
            "start": 0,
            "size": size // 512  # size is in sectors / blocks
        }
    )
    # open the device and get its loopback path
    lpath = os.path.join(
        devpath,
        devmgr.open(dev)["path"]
    )
    # mount the loopback
    mounts.MountManager(
        devmgr,
        mountpoint
    ).mount(
        mounts.Mount(
            lpath,
            index.get_module_info("Mount", "org.osbuild.fat"),
            dev,
            "/",
            options
        )
    )


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_without_options(tmpdir):
    tree, size = create_image(tmpdir)
    options = {}

    with tempfile.TemporaryDirectory(dir=tmpdir) as mountpoint:
        with host.ServiceManager() as mgr:
            with make_dev_tmpfs(tmpdir) as devpath:
                mount(mgr, devpath, tree, size, mountpoint, options)
                with open(os.path.join(mountpoint, "test"), "w", encoding="utf-8") as f:
                    f.write("should work")
                os.remove(os.path.join(mountpoint, "test"))


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_all_options(tmpdir):
    tree, size = create_image(tmpdir)
    options = {
        "readonly": True,
        "uid": 0,
        "gid": 0,
        "umask": "077",
        "shortname": "winnt"
    }
    print(options)

    with tempfile.TemporaryDirectory(dir=tmpdir) as mountpoint:
        with host.ServiceManager() as mgr:
            with make_dev_tmpfs(tmpdir) as devpath:
                mount(mgr, devpath, tree, size, mountpoint, options)

                # Check FS is read only
                with pytest.raises(OSError) as err:
                    with open(os.path.join(mountpoint, "test"), "w", encoding="utf-8") as f:
                        f.write("should not work")
                assert err.value.errno == errno.EROFS

                # Check the other options
                st = os.lstat(mountpoint)
                assert st.st_mode == 0o040700
                assert st.st_uid == 0
                assert st.st_gid == 0

                shortname_tested = False
                proc = subprocess.run("mount", capture_output=True, check=True)
                for line in proc.stdout.splitlines():
                    strline = line.decode("utf-8")
                    if mountpoint in strline:
                        assert "winnt" in strline
                        shortname_tested = True
                assert shortname_tested



"""JSON Communication

This module implements a client/server communication method based on JSON
serialization. It uses unix-domain-datagram-sockets and provides a simple
unicast message transmission.
"""


import array
import contextlib
import errno
import json
import os
import socket
from typing import Any, Optional

from .types import PathLike


class FdSet:
    """File-Descriptor Set

    This object wraps an array of file-descriptors. Unlike a normal integer
    array, this object owns the file-descriptors and therefore closes them once
    the object is released.

    File-descriptor sets are initialized once. From then one, the only allowed
    operation is to query it for information, or steal file-descriptors from
    it. If you close a set, all remaining file-descriptors are closed and
    removed from the set. It will then be an empty set.
    """

    _fds = array.array("i")

    def __init__(self, *, rawfds):
        for i in rawfds:
            if not isinstance(i, int) or i < 0:
                raise ValueError()

        self._fds = rawfds

    def __del__(self):
        self.close()

    def close(self):
        """Close All Entries

        This closes all stored file-descriptors and clears the set. Once this
        returns, the set will be empty. It is safe to call this multiple times.
        Note that a set is automatically closed when it is garbage collected.
        """

        for i in self._fds:
            if i >= 0:
                os.close(i)

        self._fds = array.array("i")

    @classmethod
    def from_list(cls, l: list):
        """Create new Set from List

        This creates a new file-descriptor set initialized to the same entries
        as in the given list. This consumes the file-descriptors. The caller
        must not assume ownership anymore.
        """

        fds = array.array("i")
        fds.fromlist(l)
        return cls(rawfds=fds)

    def __len__(self):
        return len(self._fds)

    def __getitem__(self, key: Any):
        if self._fds[key] < 0:
            raise IndexError
        return self._fds[key]

    def steal(self, key: Any):
        """Steal Entry

        Retrieve the entry at the given position, but drop it from the internal
        file-descriptor set. The caller will now own the file-descriptor and it
        can no longer be accessed through the set.

        Note that this does not reshuffle the set. All indices stay constant.
        """

        v = self[key]
        self._fds[key] = -1
        return v


class Socket(contextlib.AbstractContextManager):
    """Communication Socket

    This socket object represents a communication channel. It allows sending
    and receiving JSON-encoded messages. It uses unix-domain sequenced-packet
    sockets as underlying transport.
    """

    _socket = None
    _unlink = None

    def __init__(self, sock, unlink):
        self._socket = sock
        self._unlink = unlink

    def __del__(self):
        self.close()

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()
        return False

    @property
    def blocking(self):
        """Get the current blocking mode of the socket.

        This is related to the socket's timeout, i.e. if no time out is set
        the socket is in blocking mode; otherwise it is non-blocking.
        """
        timeout = self._socket.gettimeout()
        return timeout is not None

    @blocking.setter
    def blocking(self, value: bool):
        """Set the blocking mode of the socket."""
        if self._socket:
            self._socket.setblocking(value)
        else:
            raise RuntimeError("Tried to set blocking mode without socket.")

    def accept(self) -> Optional["Socket"]:
        """Accept a new connection on the socket.

        See python's `socket.accept` for more information.
        """

        if not self._socket:
            raise RuntimeError("Tried to accept without socket.")

        # Since, in the kernel, for AF_UNIX, new connection requests,
        # i.e. clients connecting, are directly put on the receive
        # queue of the listener socket, accept here *should* always
        # return a socket and not block, even if the client meanwhile
        # disconnected; we don't rely on that kernel behavior though
        try:
            conn, _ = self._socket.accept()
        except (socket.timeout, BlockingIOError):
            return None
        return Socket(conn, None)

    def listen(self, backlog: Optional[int] = 2**16):
        """Enable accepting of incoming connections.

        See python's `socket.listen` for details.
        """

        if not self._socket:
            raise RuntimeError("Tried to listen without socket.")

        # `Socket.listen` accepts an `int` or no argument, but not `None`
        args = [backlog] if backlog is not None else []
        self._socket.listen(*args)

    def close(self):
        """Close Socket

        Close the socket and all underlying resources. This can be called
        multiple times.
        """

        # close the socket if it is set
        if self._socket is not None:
            self._socket.close()
            self._socket = None

        # unlink the file-system entry, if pinned
        if self._unlink is not None:
            try:
                os.unlink(self._unlink[1], dir_fd=self._unlink[0])
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise

            os.close(self._unlink[0])
            self._unlink = None

    @classmethod
    def new_client(cls, connect_to: Optional[PathLike] = None):
        """Create Client

        Create a new client socket.

        Parameters
        ----------
        connect_to
            If not `None`, the client will use the specified address as the
            default destination for all send operations.
        """

        sock = None

        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)

            # Trigger an auto-bind. If you do not do this, you might end up with
            # an unbound unix socket, which cannot receive messages.
            # Alternatively, you can also set `SO_PASSCRED`, but this has
            # side-effects.
            sock.bind("")

            # Connect the socket. This has no effect other than specifying the
            # default destination for send operations.
            if connect_to is not None:
                sock.connect(os.fspath(connect_to))
        except:
            if sock is not None:
                sock.close()
            raise

        return cls(sock, None)

    @classmethod
    def new_server(cls, bind_to: PathLike):
        """Create Server

        Create a new listener socket. Returned socket is in non-blocking
        mode by default. See `blocking` property.

        Parameters
        ----------
        bind_to
            The socket-address to listen on for incoming client requests.
        """

        sock = None
        unlink = None
        path = os.path.split(bind_to)

        try:
            # We bind the socket and then open a directory-fd on the target
            # socket. This allows us to properly unlink the socket when the
            # server is closed. Note that sockets are never automatically
            # cleaned up on linux, nor can you bind to existing sockets.
            # We use a dirfd to guarantee this works even when you change
            # your mount points in-between.
            # Yeah, this is racy when mount-points change between the socket
            # creation and open. But then your entire socket creation is racy
            # as well. We do not guarantee atomicity, so you better make sure
            # you do not rely on it.
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
            sock.bind(os.fspath(bind_to))
            unlink = os.open(os.path.join(".", path[0]), os.O_CLOEXEC | os.O_PATH)
            sock.setblocking(False)
        except:
            if unlink is not None:
                os.close(unlink)
            if sock is not None:
                sock.close()
            raise

        return cls(sock, (unlink, path[1]))

    @classmethod
    def new_pair(cls, *, blocking=True):
        """Create a connected socket pair

        Create a pair of connected sockets and return both as a tuple.

        Parameters
        ----------
        blocking
            The blocking mode for the socket pair.
        """
        a, b = socket.socketpair(socket.AF_UNIX, socket.SOCK_SEQPACKET)

        a.setblocking(blocking)
        b.setblocking(blocking)

        return cls(a, None), cls(b, None)

    @classmethod
    def new_from_fd(cls, fd: int, *, blocking=True, close_fd=True):
        """Create a socket for an existing file descriptor

        Duplicate the file descriptor and return a `Socket` for it.
        The blocking mode can be set via `blocking`. If `close_fd`
        is True (the default) `fd` will be closed.

        Parameters
        ----------
        fd
            The file descriptor to use.
        blocking
            The blocking mode for the socket pair.
        """
        sock = socket.fromfd(fd, socket.AF_UNIX, socket.SOCK_SEQPACKET)
        sock.setblocking(blocking)
        if close_fd:
            os.close(fd)
        return cls(sock, None)

    def fileno(self) -> int:
        assert self._socket is not None
        return self._socket.fileno()

    def recv(self):
        """Receive a Message

        This receives the next pending message from the socket. This operation
        is synchronous.

        A tuple consisting of the deserialized message payload, the auxiliary
        file-descriptor set, and the socket-address of the sender is returned.

        In case the peer closed the connection, A tuple of `None` values is
        returned.
        """

        # On `SOCK_SEQPACKET`, packets might be arbitrarily sized. There is no
        # hard-coded upper limit, since it is only restricted by the size of
        # the kernel write buffer on sockets (which itself can be modified via
        # sysctl). The only real maximum is probably something like 2^31-1,
        # since that is the maximum of that sysctl datatype.
        # Anyway, `MSG_TRUNC+MSG_PEEK` usually allows us to easily peek at the
        # incoming buffer. Unfortunately, the python `recvmsg()` wrapper
        # discards the return code and we cannot use that. Instead, we simply
        # loop until we know the size. This is slightly awkward, but seems fine
        # as long as you do not put this into a hot-path.
        size = 4096
        while True:
            peek = self._socket.recvmsg(size, 0, socket.MSG_PEEK)
            if not peek[0]:
                # Connection was closed
                return None, None, None
            if not (peek[2] & socket.MSG_TRUNC):
                break
            size *= 2

        # Fetch a packet from the socket. On linux, the maximum SCM_RIGHTS array
        # size is hard-coded to 253. This allows us to size the ancillary buffer
        # big enough to receive any possible message.
        fds = array.array("i")
        msg = self._socket.recvmsg(size, socket.CMSG_LEN(253 * fds.itemsize))

        # First thing we do is always to fetch the CMSG FDs into an FdSet. This
        # guarantees that we do not leak FDs in case the message handling fails
        # for other reasons.
        for level, ty, data in msg[1]:
            if level == socket.SOL_SOCKET and ty == socket.SCM_RIGHTS:
                assert len(data) % fds.itemsize == 0
                fds.frombytes(data)
        fdset = FdSet(rawfds=fds)

        # Check the returned message flags. If the message was truncated, we
        # have to discard it. This shouldn't happen, but there is no harm in
        # handling it. However, `CTRUNC` can happen, since it is also triggered
        # when LSMs reject FD transmission. Treat it the same as a parser error.
        flags = msg[2]
        if flags & (socket.MSG_TRUNC | socket.MSG_CTRUNC):
            raise BufferError

        try:
            payload = json.loads(msg[0])
        except json.JSONDecodeError as e:
            raise BufferError from e

        return (payload, fdset, msg[3])

    def send(self, payload: object, *, fds: Optional[list] = None):
        """Send Message

        Send a new message via this socket. This operation is synchronous. The
        maximum message size depends on the configured send-buffer on the
        socket. An `OSError` with `EMSGSIZE` is raised when it is exceeded.

        Parameters
        ----------
        payload
            A python object to serialize as JSON and send via this socket. See
            `json.dump()` for details about the serialization involved.
        destination
            The destination to send to. If `None`, the default destination is
            used (if none is set, this will raise an `OSError`).
        fds
            A list of file-descriptors to send with the message.

        Raises
        ------
        OSError
            If the socket cannot be written, a matching `OSError` is raised.
        TypeError
            If the payload cannot be serialized, a type error is raised.
        """

        if not self._socket:
            raise RuntimeError("Tried to send without socket.")

        serialized = json.dumps(payload).encode()
        cmsg = []
        if fds:
            cmsg.append((socket.SOL_SOCKET, socket.SCM_RIGHTS, array.array("i", fds)))

        n = self._socket.sendmsg([serialized], cmsg, 0)
        assert n == len(serialized)

    def send_and_recv(self, payload: object, *, fds: Optional[list] = None):
        """Send a message and wait for a reply

        This is a convenience helper that combines `send` and `recv`.
        See the individual methods for details about the parameters.
        """

        self.send(payload, fds=fds)
        return self.recv()



#
# Run `pylint` on all python sources.
#

import os
import subprocess

import pytest

from .. import test


@pytest.fixture(name="source_files")
def discover_source_files():

    # Evaluate whether `path` is a python file. This assumes that the file is
    # opened by the caller, anyway, hence it might invoke `file` to detect the
    # file type if in doubt.
    def is_python_script(path):
        if path.endswith(".py"):
            return True

        mime_valid = [
            "text/x-python",
            "text/x-script.python",
        ]

        mime_file = subprocess.check_output(
            [
                "file",
                "-bi",
                "--",
                path,
            ],
            encoding="utf-8",
        )

        return any(mime_file.startswith(v) for v in mime_valid)

    # Enumerate all repository files. This will invoke `git ls-tree` to list
    # all files. This also requires a temporary change of directory, since git
    # operates on the current directory and does not allow path arguments.
    def list_files(path):
        cwd = os.getcwd()
        os.chdir(path)
        files = subprocess.check_output(
            [
                "git",
                "ls-tree",
                "-rz",
                "--full-tree",
                "--name-only",
                "HEAD",
            ],
            encoding="utf-8",
        )
        os.chdir(cwd)

        files = files.split('\x00')
        return (os.path.join(path, f) for f in files)

    if not test.TestBase.have_test_checkout():
        pytest.skip("no test-checkout access")

    files = list_files(test.TestBase.locate_test_checkout())
    files = filter(is_python_script, files)
    return list(files)


def test_pylint(source_files):
    #
    # Run `pylint` on all python sources. We simply use `find` to locate
    # all `*.py` files, and then manually select the reverse-domain named
    # modules we have.
    #

    r = subprocess.run(["pylint"] + source_files, check=False)
    if r.returncode != 0:
        pytest.fail("pylint issues detected")


@pytest.mark.skipif(not test.TestBase.have_mypy(), reason="mypy not available")
def test_mypy():
    #
    # Run `mypy` on osbuild sources.
    #

    r = subprocess.run(["mypy", "osbuild/"], check=False)
    if r.returncode != 0:
        pytest.fail("mypy issues detected")


@pytest.mark.skipif(not test.TestBase.have_isort(), reason="isort not available")
def test_isort(source_files):
    #
    # Run `isort` on all python sources. We simply use `find` to locate
    # all `*.py` files, and then manually select the reverse-domain named
    # modules we have.
    #

    r = subprocess.run(["isort", "--check", "--diff"] + source_files, check=False)
    if r.returncode != 0:
        pytest.fail("isort issues detected")


@pytest.mark.skipif(not test.TestBase.have_autopep8(), reason="autopep8 not available")
def test_autopep8(source_files):
    r = subprocess.run(["autopep8-3", "--diff", "--exit-code"] + source_files, check=False)
    if r.returncode != 0:
        pytest.fail("autopep8 has detected changes (see diff)")



#
# Tests for the `osbuild.util.linux` module.
#

import os
import subprocess
import tempfile

import pytest

from osbuild.util import linux

from .. import test


@pytest.fixture(name="tmpdir")
def tmpdir_fixture():
    with tempfile.TemporaryDirectory(dir="/var/tmp") as tmp:
        yield tmp


@pytest.mark.skipif(not test.TestBase.can_modify_immutable("/var/tmp"), reason="root-only")
def test_ioctl_get_immutable(tmpdir):
    #
    # Test the `ioctl_get_immutable()` helper and make sure it works
    # as intended.
    #

    with open(f"{tmpdir}/immutable", "x", encoding="utf8") as f:
        assert not linux.ioctl_get_immutable(f.fileno())


@pytest.mark.skipif(not test.TestBase.can_modify_immutable("/var/tmp"), reason="root-only")
def test_ioctl_toggle_immutable(tmpdir):
    #
    # Test the `ioctl_toggle_immutable()` helper and make sure it works
    # as intended.
    #

    with open(f"{tmpdir}/immutable", "x", encoding="utf8") as f:
        # Check the file is mutable by default and if we clear it again.
        assert not linux.ioctl_get_immutable(f.fileno())
        linux.ioctl_toggle_immutable(f.fileno(), False)
        assert not linux.ioctl_get_immutable(f.fileno())

        # Set immutable and check for it. Try again to verify with flag set.
        linux.ioctl_toggle_immutable(f.fileno(), True)
        assert linux.ioctl_get_immutable(f.fileno())
        linux.ioctl_toggle_immutable(f.fileno(), True)
        assert linux.ioctl_get_immutable(f.fileno())

        # Verify immutable files cannot be unlinked.
        with pytest.raises(OSError):
            os.unlink(f"{tmpdir}/immutable")

        # Check again that clearing the flag works.
        linux.ioctl_toggle_immutable(f.fileno(), False)
        assert not linux.ioctl_get_immutable(f.fileno())

        # This time, check that we actually set the same flag as `chattr`.
        subprocess.run(["chattr", "+i",
                        f"{tmpdir}/immutable"], check=True)
        assert linux.ioctl_get_immutable(f.fileno())

        # Same for clearing it.
        subprocess.run(["chattr", "-i",
                        f"{tmpdir}/immutable"], check=True)
        assert not linux.ioctl_get_immutable(f.fileno())

        # Verify we can unlink the file again, once the flag is cleared.
        os.unlink(f"{tmpdir}/immutable")


@pytest.mark.skipif(not linux.cap_is_supported(), reason="no support for capabilities")
def test_capabilities():
    #
    # Test the capability related utility functions
    #

    lib = linux.LibCap.get_default()
    assert lib
    l2 = linux.LibCap.get_default()
    assert lib is l2

    assert linux.cap_is_supported()

    assert linux.cap_is_supported("CAP_MAC_ADMIN")

    val = lib.from_name("CAP_MAC_ADMIN")
    assert val >= 0

    name = lib.to_name(val)
    assert name == "CAP_MAC_ADMIN"

    assert not linux.cap_is_supported("CAP_GICMO")
    with pytest.raises(OSError):
        lib.from_name("CAP_GICMO")


def test_fcntl_flock():
    #
    # This tests the `linux.fcntl_flock()` file-locking helper. Note
    # that file-locks are on the open-file-description, so they are shared
    # between dupped file-descriptors. We explicitly create a separate
    # file-description via `/proc/self/fd/`.
    #

    with tempfile.TemporaryFile() as f:
        fd1 = f.fileno()
        fd2 = os.open(os.path.join("/proc/self/fd/", str(fd1)), os.O_RDWR | os.O_CLOEXEC)

        # Test: unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: write-lock + unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: read-lock1 + read-lock2 + unlock1 + unlock2
        linux.fcntl_flock(fd1, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd2, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)
        linux.fcntl_flock(fd2, linux.fcntl.F_UNLCK)

        # Test: write-lock1 + write-lock2 + unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        with pytest.raises(BlockingIOError):
            linux.fcntl_flock(fd2, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: write-lock1 + read-lock2 + unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        with pytest.raises(BlockingIOError):
            linux.fcntl_flock(fd2, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: read-lock1 + write-lock2 + unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_RDLCK)
        with pytest.raises(BlockingIOError):
            linux.fcntl_flock(fd2, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: write-lock1 + read-lock1 + read-lock2 + unlock
        linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd2, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Test: read-lock1 + read-lock2 + write-lock1 + unlock1 + unlock2
        linux.fcntl_flock(fd1, linux.fcntl.F_RDLCK)
        linux.fcntl_flock(fd2, linux.fcntl.F_RDLCK)
        with pytest.raises(BlockingIOError):
            linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)
        linux.fcntl_flock(fd2, linux.fcntl.F_UNLCK)

        # Test: write-lock3 + write-lock1 + close3 + write-lock1 + unlock1
        fd3 = os.open(os.path.join("/proc/self/fd/", str(fd1)), os.O_RDWR | os.O_CLOEXEC)
        linux.fcntl_flock(fd3, linux.fcntl.F_WRLCK)
        with pytest.raises(BlockingIOError):
            linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        os.close(fd3)
        linux.fcntl_flock(fd1, linux.fcntl.F_WRLCK)
        linux.fcntl_flock(fd1, linux.fcntl.F_UNLCK)

        # Cleanup
        os.close(fd2)


def test_libc():
    #
    # Test that the Libc class can be instantiated and provides a suitable
    # default singleton. Verify the expected interfaces exist (though tests
    # for them are separate).
    #

    libc0 = linux.Libc.make()
    libc1 = linux.Libc.default()

    assert libc0 is not libc1
    assert libc1 is linux.Libc.default()

    assert libc0.AT_FDCWD
    assert libc0.RENAME_EXCHANGE
    assert libc0.RENAME_NOREPLACE
    assert libc0.RENAME_WHITEOUT
    assert libc0.renameat2


def test_libc_renameat2_errcheck():
    #
    # Verify the `renameat(2)` system call on `Libc` correctly turns errors into
    # python exceptions.
    #

    libc = linux.Libc.default()

    with pytest.raises(OSError):
        libc.renameat2(oldpath=b"", newpath=b"")


def test_libc_renameat2_exchange(tmpdir):
    #
    # Verify the `renameat(2)` system call on `Libc` with the
    # `RENAME_EXCHANGE` flag. This swaps two files atomically.
    #

    libc = linux.Libc.default()

    with open(f"{tmpdir}/foo", "x", encoding="utf8") as f:
        f.write("foo")
    with open(f"{tmpdir}/bar", "x", encoding="utf8") as f:
        f.write("bar")

    libc.renameat2(
        oldpath=f"{tmpdir}/foo".encode(),
        newpath=f"{tmpdir}/bar".encode(),
        flags=linux.Libc.RENAME_EXCHANGE,
    )

    with open(f"{tmpdir}/foo", "r", encoding="utf8") as f:
        assert f.read() == "bar"
    with open(f"{tmpdir}/bar", "r", encoding="utf8") as f:
        assert f.read() == "foo"


def test_proc_boot_id():
    #
    # Test the `proc_boot_id()` function which reads the current boot-id
    # from the kernel. Make sure it is a valid UUID and also consistent on
    # repeated queries.
    #

    bootid = linux.proc_boot_id("test")
    assert len(bootid.hex) == 32
    assert bootid.version == 4

    bootid2 = linux.proc_boot_id("test")
    assert bootid.int == bootid2.int

    bootid3 = linux.proc_boot_id("foobar")
    assert bootid.int != bootid3.int



#
# Basic tests for a collection of osbuild modules.
#

import json
import os
import pathlib
import sys
import tempfile
import unittest

import osbuild
import osbuild.meta
from osbuild.monitor import NullMonitor
from osbuild.objectstore import ObjectStore
from osbuild.pipeline import Manifest, Runner

from .. import test


def names(*lst):
    return [x.name for x in lst]


class MockStore:
    def __init__(self) -> None:
        self.have = set()

    def contains(self, pipeline_id):
        return pipeline_id in self.have


class TestDescriptions(unittest.TestCase):

    @unittest.skipUnless(test.TestBase.can_bind_mount(), "root-only")
    def test_stage_run(self):
        index = osbuild.meta.Index(os.curdir)
        info = index.get_module_info("Stage", "org.osbuild.noop")
        stage = osbuild.Stage(info, {}, None, None, {}, None)

        with tempfile.TemporaryDirectory() as tmpdir:

            storedir = pathlib.Path(tmpdir, "store")
            root = pathlib.Path("/")
            runner = Runner(index.detect_host_runner())
            monitor = NullMonitor(sys.stderr.fileno())
            libdir = os.path.abspath(os.curdir)
            store = ObjectStore(storedir)

            with ObjectStore(storedir) as store:
                data = store.new(stage.id)
                res = stage.run(data, runner, root, store, monitor, libdir)

        self.assertEqual(res.success, True)
        self.assertEqual(res.id, stage.id)

    def test_manifest(self):
        index = osbuild.meta.Index(os.curdir)

        info = index.get_module_info("Stage", "org.osbuild.noop")

        manifest = Manifest()

        # each pipeline gets a noop stage with different
        # options so that their ids are different
        build = manifest.add_pipeline("build", None, None)
        build.add_stage(info, {"option": 1})

        tree = manifest.add_pipeline("tree",
                                     "org.osbuild.linux",
                                     build.id)
        tree.add_stage(info, {"option": 2})

        assembler = manifest.add_pipeline("assembler",
                                          "org.osbuild.inux",
                                          build.id)
        assembler.add_stage(info, {"option": 3})

        self.assertEqual(len(manifest.pipelines), 3)

        self.assertIn("build", manifest.pipelines)
        self.assertIn("tree", manifest.pipelines)
        self.assertIn("assembler", manifest.pipelines)

        self.assertIn("build", manifest)
        self.assertIn("tree", manifest)
        self.assertIn("assembler", manifest)

        # make sure the order is correct
        lst = ["build", "tree", "assembler"]
        for a, b in zip(manifest.pipelines, lst):
            self.assertEqual(a, b)

        for a, b in zip(manifest, [build, tree, assembler]):
            self.assertEqual(a.name, b.name)

        # check we get exceptions on unknown names
        with self.assertRaises(KeyError):
            _ = manifest.pipelines["foo"]

        with self.assertRaises(KeyError):
            _ = manifest["foo"]

        # check helper functions
        #  access by name
        check = manifest["build"]
        self.assertEqual(build.name, check.name)

        check = manifest["tree"]
        self.assertEqual(tree.name, check.name)

        check = manifest["assembler"]
        self.assertEqual(assembler.name, check.name)

        #  `None` return for unknown items
        check = manifest.get("foo")
        self.assertIsNone(check)

        #  id based access
        for i in [build, tree, assembler]:
            check = manifest[i.id]
            self.assertEqual(i.name, check.name)

    # pylint: disable=too-many-statements
    def test_on_demand(self):
        index = osbuild.meta.Index(os.curdir)
        host_runner = Runner(index.detect_host_runner())
        runner = Runner(index.detect_runner("org.osbuild.linux"))

        manifest = Manifest()
        noop = index.get_module_info("Stage", "org.osbuild.noop")
        noip = index.get_module_info("Input", "org.osbuild.noop")

        # the shared build pipeline
        build = manifest.add_pipeline("build", host_runner, None)
        build.add_stage(noop, {"option": 1})

        # a pipeline simulating some intermediate artefact
        # that other pipeline need as dependency
        dep = manifest.add_pipeline("dep",
                                    runner,
                                    build.id)

        dep.add_stage(noop, {"option": 2})

        # a pipeline that is not linked to the "main"
        # assembler artefact and thus should normally
        # not be built unless explicitly requested
        # has an input that depends on `dep`
        ul = manifest.add_pipeline("unlinked",
                                   runner,
                                   build.id)

        stage = ul.add_stage(noop, {"option": 3})
        ip = stage.add_input("dep", noip, "org.osbuild.pipeline")
        ip.add_reference(dep.id)

        # the main os root file system
        rootfs = manifest.add_pipeline("rootfs",
                                       runner,
                                       build.id)
        stage = rootfs.add_stage(noop, {"option": 4})

        # the main raw image artefact, depending on "dep" and
        # "rootfs"
        image = manifest.add_pipeline("image",
                                      runner,
                                      build.id)

        stage = image.add_stage(noop, {"option": 5})
        ip = stage.add_input("dep", noip, "org.osbuild.pipeline")
        ip.add_reference(dep.id)

        stage = image.add_stage(noop, {"option": 6})

        #  a stage using the rootfs as input (named 'image')
        ip = stage.add_input("image", noip, "org.osbuild.pipeline")
        ip.add_reference(rootfs.id)

        # some compression of the image, like a qcow2
        qcow2 = manifest.add_pipeline("qcow2",
                                      runner,
                                      build.id)
        stage = qcow2.add_stage(noop, {"option": 7})
        ip = stage.add_input("image", noip, "org.osbuild.pipeline")
        ip.add_reference(image.id)

        fmt = index.get_format_info("osbuild.formats.v2").module
        self.assertIsNotNone(fmt)
        print(json.dumps(fmt.describe(manifest), indent=2))

        # The pipeline graph in the manifest with dependencies:
        # ├─╼ build
        # ├─╼ dep
        # │ └ build
        # ├─╼ unlinked
        # │ ├ build
        # │ └ dep
        # ├─╼ rootfs
        # │ └ build
        # ├─╼ image
        # │ ├ build
        # │ ├ dep
        # │ └ rootfs
        # └─╼ qcow2
        #   ├ build
        #   └ image

        store = MockStore()

        # check an empty input leads to an empty list
        res = manifest.depsolve(store, [])
        assert not res

        # the build pipeline should resolve to just itself
        res = manifest.depsolve(store, names(build))
        assert res == names(build)

        # if we build the 'unlinked' pipeline, we get it
        # and its dependencies, dep and build
        res = manifest.depsolve(store, names(ul))
        assert res == names(build, dep, ul)

        # building image with nothing in the store should
        # result in all pipelines but 'unlinked'
        res = manifest.depsolve(store, names(image))
        assert res == names(build, rootfs, dep, image)

        # ensure the order of inputs is preserved during
        # the depsolving so that we build things in the
        # same way they were requested
        res = manifest.depsolve(store, names(ul, image))
        assert res == names(build, dep, ul, rootfs, image)

        res = manifest.depsolve(store, names(image, ul))
        assert res == names(build, rootfs, dep, image, ul)

        # if we have the 'dep' dependency in the store,
        # we should be not be building that
        store.have.add(dep.id)
        res = manifest.depsolve(store, names(image))
        assert res == names(build, rootfs, image)

        # if we only have the build pipeline in the
        # store we should not build that
        store.have.clear()
        store.have.add(build.id)
        res = manifest.depsolve(store, names(image))
        assert res == names(rootfs, dep, image)

        # if we have the final artefact in the store,
        # nothing should be built at all
        store.have.clear()
        store.have.add(image.id)
        res = manifest.depsolve(store, names(image))
        assert not res

        # we have a checkpoint of the stage in the image
        # pipeline with the `dep` dependency, so that
        # it effectively only depends on `rootfs`
        store.have.clear()
        store.have.add(image.stages[0].id)
        res = manifest.depsolve(store, names(image))
        assert res == names(build, rootfs, image)

    def check_moduleinfo(self, version):
        index = osbuild.meta.Index(os.curdir)

        modules = []
        for klass in ("Assembler", "Device", "Input", "Mount", "Source", "Stage"):
            mods = index.list_modules_for_class(klass)
            modules += [(klass, module) for module in mods]

        self.assertTrue(modules)

        for module in modules:
            klass, name = module
            try:
                info = osbuild.meta.ModuleInfo.load(os.curdir, klass, name)
                schema = osbuild.meta.Schema(info.get_schema(version), name)
                res = schema.check()
                if not res:
                    err = "SCHEMA: " + json.dumps(schema.data, indent=2) + "\n"
                    err += "\n  ".join(str(e) for e in res)
                    self.fail(str(res) + "\n  " + err)
            except json.decoder.JSONDecodeError as e:
                msg = f"{klass} '{name}' has invalid STAGE_OPTS\n\t" + str(e)
                self.fail(msg)

    def test_moduleinfo(self):
        for version in ["1", "2"]:
            with self.subTest(version=version):
                self.check_moduleinfo(version)



#
# Test for the build root
#

import os
import pathlib
import sys
from tempfile import TemporaryDirectory

import pytest

import osbuild.meta
from osbuild.buildroot import BuildRoot
from osbuild.monitor import LogMonitor, NullMonitor
from osbuild.util import linux

from ..test import TestBase


@pytest.fixture(name="tempdir")
def tempdir_fixture():
    with TemporaryDirectory(prefix="lvm2-") as tmp:
        yield tmp


@pytest.fixture(name="runner")
def runner_fixture():
    meta = osbuild.meta.Index(os.curdir)
    runner = meta.detect_host_runner()
    return runner.path


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_basic(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        r = root.run(["/usr/bin/true"], monitor)
        assert r.returncode == 0

        # Test we can use `.run` multiple times
        r = root.run(["/usr/bin/true"], monitor)
        assert r.returncode == 0

        r = root.run(["/usr/bin/false"], monitor)
        assert r.returncode != 0


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_runner_fail(tempdir):
    runner = "org.osbuild.nonexistantrunner"
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    logfile = os.path.join(tempdir, "log.txt")

    with BuildRoot("/", runner, libdir, var) as root, \
            open(logfile, "w", encoding="utf8") as log:

        monitor = LogMonitor(log.fileno())

        r = root.run(["/usr/bin/true"], monitor)

    assert r.returncode == 1
    with open(logfile, encoding="utf8") as f:
        log = f.read()
    assert log
    assert r.output
    assert log == r.output


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_output(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    data = "42. cats are superior to dogs"

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        r = root.run(["/usr/bin/echo", data], monitor)
        assert r.returncode == 0

    assert data in r.output.strip()


@pytest.mark.skipif(not TestBase.have_test_data(), reason="no test-data access")
@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_bind_mounts(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    rw_data = pathlib.Path(tempdir, "data")
    rw_data.mkdir()

    scripts = os.path.join(TestBase.locate_test_data(), "scripts")

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        ro_binds = [f"{scripts}:/scripts"]

        cmd = ["/scripts/mount_flags.py",
               "/scripts",
               "ro"]

        r = root.run(cmd, monitor, readonly_binds=ro_binds)
        assert r.returncode == 0

        cmd = ["/scripts/mount_flags.py",
               "/rw-data",
               "ro"]

        binds = [f"{rw_data}:/rw-data"]
        r = root.run(cmd, monitor, binds=binds, readonly_binds=ro_binds)
        assert r.returncode == 1


@pytest.mark.skipif(not TestBase.have_test_data(), reason="no test-data access")
@pytest.mark.skipif(not os.path.exists("/sys/fs/selinux"), reason="no SELinux")
def test_selinuxfs_ro(tempdir, runner):
    # /sys/fs/selinux must never be writable in the container
    # because RPM and other tools must not assume the policy
    # of the host is the valid policy

    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    scripts = os.path.join(TestBase.locate_test_data(), "scripts")

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        ro_binds = [f"{scripts}:/scripts"]

        cmd = ["/scripts/mount_flags.py",
               "/sys/fs/selinux",
               "ro"]

        r = root.run(cmd, monitor, readonly_binds=ro_binds)
        assert r.returncode == 0


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_proc_overrides(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    cmdline = "is-this-the-real-world"

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        root.proc.cmdline = cmdline

        r = root.run(["cat", "/proc/cmdline"], monitor)
        assert r.returncode == 0
        assert cmdline in r.output.strip()


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_timeout(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    monitor = NullMonitor(sys.stderr.fileno())

    with BuildRoot("/", runner, libdir, var) as root:

        root.run(["/bin/sleep", "1"], monitor, timeout=2)

        with pytest.raises(TimeoutError):
            root.run(["/bin/sleep", "1"], monitor, timeout=0.1)

        with pytest.raises(TimeoutError):
            root.run(["/bin/sleep", "1"], monitor, timeout=0.1)


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_env_isolation(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    monitor = NullMonitor(sys.stderr.fileno())

    ipc = pathlib.Path(tempdir, "ipc")
    ipc.mkdir()

    # Set some env variable to make sure it is not leaked into
    # the container
    os.environ["OSBUILD_TEST_ENV_ISOLATION"] = "42"

    with BuildRoot("/", runner, libdir, var) as root:
        cmd = ["/bin/sh", "-c", "/usr/bin/env > /ipc/env.txt"]
        r = root.run(cmd, monitor, binds=[f"{ipc}:/ipc"])

    assert r.returncode == 0
    with open(os.path.join(ipc, "env.txt"), encoding="utf8") as f:
        data = f.read().strip()
    assert data
    have = dict(map(lambda x: x.split("=", 1), data.split("\n")))

    allowed = [
        "_",      # added by `env` itself
        "container",
        "LC_CTYPE",
        "PATH",
        "PWD",
        "PYTHONPATH",
        "PYTHONUNBUFFERED",
        "SHLVL",  # added by the shell wrapper
        "TERM",
    ]

    for k in have:
        assert k in allowed


@pytest.mark.skipif(not TestBase.can_bind_mount(), reason="root only")
def test_caps(tempdir, runner):
    libdir = os.path.abspath(os.curdir)
    var = pathlib.Path(tempdir, "var")
    var.mkdir()

    ipc = pathlib.Path(tempdir, "ipc")
    ipc.mkdir()

    monitor = NullMonitor(sys.stderr.fileno())
    with BuildRoot("/", runner, libdir, var) as root:

        def run_and_get_caps():
            cmd = ["/bin/sh", "-c", "cat /proc/self/status > /ipc/status"]
            r = root.run(cmd, monitor, binds=[f"{ipc}:/ipc"])

            assert r.returncode == 0
            with open(os.path.join(ipc, "status"), encoding="utf8") as f:
                data = f.readlines()
            assert data

            print(data)
            perm = list(filter(lambda x: x.startswith("CapEff"), data))

            assert perm and len(perm) == 1
            perm = perm[0]

            perm = perm[7:].strip()  # strip "CapEff"
            print(perm)

            caps = linux.cap_mask_to_set(int(perm, base=16))
            return caps

        # check case of `BuildRoot.caps` is `None`, i.e. don't drop capabilities,
        # thus the effective capabilities should be the bounding set
        assert root.caps is None

        bound_set = linux.cap_bound_set()

        caps = run_and_get_caps()
        assert caps == bound_set

        # drop everything but `CAP_SYS_ADMIN`
        assert "CAP_SYS_ADMIN" in bound_set

        enable = set(["CAP_SYS_ADMIN"])
        disable = bound_set - enable

        root.caps = enable

        caps = run_and_get_caps()

        for e in enable:
            assert e in caps
        for d in disable:
            assert d not in caps



"""Mount utilities
"""

import contextlib
import subprocess


def mount(source, target, bind=True, ro=True, private=True, mode="0755"):
    options = []
    if ro:
        options += ["ro"]
    if mode:
        options += [mode]

    args = []
    if bind:
        args += ["--rbind"]
    if private:
        args += ["--make-rprivate"]
    if options:
        args += ["-o", ",".join(options)]

    r = subprocess.run(["mount"] + args + [source, target],
                       stderr=subprocess.STDOUT,
                       stdout=subprocess.PIPE,
                       encoding="utf-8",
                       check=False)

    if r.returncode != 0:
        code = r.returncode
        msg = r.stdout.strip()
        raise RuntimeError(f"{msg} (code: {code})")


def umount(target, lazy=False):
    args = []
    if lazy:
        args += ["--lazy"]
    subprocess.run(["sync", "-f", target], check=True)
    subprocess.run(["umount", "-R"] + args + [target], check=True)


class MountGuard(contextlib.AbstractContextManager):
    def __init__(self):
        self.mounts = []

    def mount(self, source, target, bind=True, ro=False, mode="0755"):
        options = []
        if bind:
            options += ["bind"]
        if ro:
            options += ["ro"]
        if mode:
            options += [mode]

        args = ["--make-private"]
        if options:
            args += ["-o", ",".join(options)]

        subprocess.run(["mount"] + args + [source, target], check=True)
        self.mounts += [{"source": source, "target": target}]

    def umount(self):

        while self.mounts:
            mnt = self.mounts.pop()  # FILO: get the last mount
            target = mnt["target"]
            # The sync should in theory not be needed but in rare
            # cases `target is busy` error has been spotted.
            # Calling  `sync` does not hurt so we keep it for now.
            subprocess.run(["sync", "-f", target], check=True)
            subprocess.run(["umount", target], check=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.umount()



import os
from tempfile import TemporaryDirectory

import pytest

import osbuild
import osbuild.meta


@pytest.fixture(name="tempdir")
def tempdir_fixture():
    with TemporaryDirectory(prefix="meta-") as tmp:
        yield tmp


def create_runners(path, distro: str, base: str, versions):
    """Create runner"""
    runners = []

    basename = f"{distro}{base}"
    basepath = os.path.join(path, basename)
    with open(basepath, "x", encoding="utf-8") as f:
        f.write("#!/bin/bash")
        runners.append(basename)

    for x in versions:
        name = f"{distro}{x}"
        link = os.path.join(path, name)
        os.symlink(basepath, link)
        runners.append(name)

    return runners


def test_parse_name():
    table = {
        "arch": {
            "distro": "arch",
            "version": 0,
        },
        "fedora30": {
            "distro": "fedora",
            "version": 30,
        },
        "rhel7": {
            "distro": "rhel",
            "version": 7,
        },
        "ubuntu1804": {
            "distro": "ubuntu",
            "version": 1804,
        }
    }

    for name, want in table.items():
        d, v = osbuild.meta.RunnerInfo.parse_name(name)

        assert d == want["distro"]
        assert v == want["version"]


def test_runner_detection(tempdir):

    runners = os.path.join(tempdir, "runners")
    os.makedirs(runners)

    table = {
        "arch": {
            "base": "",
            "versions": [],
            "check": {"": 0},
            "fail": []
        },
        "fedora": {
            "base": 30,
            "versions": list(range(31, 40)),
            "check": {40: 39, 31: 31, 35: 35},
            "fail": [29],
        },
        "ubuntu": {
            "base": 1810,
            "versions": [1904, 1910, 2004],
            "check": {2010: 2004, 1912: 1910},
            "fail": [1804],
        },
        "rhel": {
            "base": 90,
            "versions":  [91, 92, 93],
            "check": {94: 93},
        },
        "future": {
            "base": 100,
            "versions":  [101, 102, 103],
            "check": {110: 103},
        }
    }

    want_all = []
    for distro, info in table.items():
        base = info["base"] or 0
        versions = info["versions"]
        want = create_runners(runners, distro, str(base), versions)
        meta = osbuild.meta.Index(tempdir)
        have = meta.list_runners(distro)
        assert len(want) == len(have)
        want_all += want

        for v in [base] + versions:
            name = f"{distro}{v}"
            runner = meta.detect_runner(name)
            assert runner
            assert runner.distro == distro
            assert runner.version == v

        for v, t in info["check"].items():
            name = f"{distro}{v}"
            runner = meta.detect_runner(name)
            assert runner
            assert runner.distro == distro
            assert runner.version == t

        for v in info.get("fail", []):
            name = f"{distro}{v}"
            with pytest.raises(ValueError):
                runner = meta.detect_runner(name)

    have = meta.list_runners()
    assert len(have) == len(want_all)


def test_runner_sorting(tempdir):

    runners = os.path.join(tempdir, "runners")
    os.makedirs(runners)

    table = {
        "A": {
            "base": 1,
            "versions": [2, 3]
        },
        "B": {
            "base": 1,
            "versions": [2, 3]
        }
    }

    for distro, info in table.items():
        base = info["base"] or 0
        versions = info["versions"]
        create_runners(runners, distro, str(base), versions)

    meta = osbuild.meta.Index(tempdir)
    have = meta.list_runners()

    names = [
        f"{i.distro}{i.version}" for i in have
    ]

    assert names == ["A1", "A2", "A3", "B1", "B2", "B3"]


def test_schema():
    schema = osbuild.meta.Schema(None)
    assert not schema

    schema = osbuild.meta.Schema({"type": "bool"})  # should be 'boolean'
    assert not schema.check().valid
    assert not schema

    schema = osbuild.meta.Schema({"type": "array", "minItems": 3})
    assert schema.check().valid
    assert schema

    res = schema.validate([1, 2])
    assert not res
    res = schema.validate([1, 2, 3])
    assert res



"""Path handling utility functions"""
import errno
import os
import os.path
from typing import Optional

from .ctx import suppress_oserror


def clamp_mtime(path: str, start: int, to: int):
    """Clamp all modification times of 'path'

    Set the mtime of 'path' to 'to' if it is greater or equal to 'start'.
    If 'to' is None, the mtime is set to the current time.
    """

    times = (to, to)

    def fix_utime(path, dfd: Optional[int] = None):
        sb = os.stat(path, dir_fd=dfd, follow_symlinks=False)
        if sb.st_mtime < start:
            return

        # We might get a permission error when the immutable flag is set;
        # since there is nothing much we can do, we just ignore it
        with suppress_oserror(errno.EPERM):
            os.utime(path, times, dir_fd=dfd, follow_symlinks=False)

    fix_utime(path)

    for _, dirs, files, dfd in os.fwalk(path):
        for f in dirs + files:
            fix_utime(f, dfd)


def in_tree(path: str, tree: str, must_exist: bool = False) -> bool:
    """Return whether the canonical location of 'path' is under 'tree'.
    If 'must_exist' is True, the file must also exist for the check to succeed.
    """
    path = os.path.abspath(path)
    if path.startswith(tree):
        return not must_exist or os.path.exists(path)
    return False



#
# Tests for the 'osbuild.objectstore' module.
#

import contextlib
import os
import tempfile
from pathlib import Path

import pytest

from osbuild import objectstore

from .. import test


def store_path(store: objectstore.ObjectStore, ref: str, path: str) -> bool:
    obj = store.get(ref)
    if not obj:
        return False
    return os.path.exists(os.path.join(obj, path))


@pytest.fixture(name="tmpdir")
def tmpdir_fixture():
    with tempfile.TemporaryDirectory() as tmp:
        yield tmp


@pytest.fixture(name="object_store")
def store_fixture():
    with tempfile.TemporaryDirectory(
        prefix="osbuild-test-",
        dir="/var/tmp",
    ) as tmp:
        with objectstore.ObjectStore(tmp) as store:
            yield store


def test_basic(object_store):
    object_store.maximum_size = 1024*1024*1024

    # No objects or references should be in the store
    assert len(os.listdir(object_store.objects)) == 0

    tree = object_store.new("a")

    # new object should be in write mode
    assert tree.mode == objectstore.Object.Mode.WRITE

    p = Path(tree, "A")
    p.touch()

    tree.finalize()  # put the object into READ mode
    assert tree.mode == objectstore.Object.Mode.READ

    # commit makes a copy, if space
    object_store.commit(tree, "a")
    assert store_path(object_store, "a", "A")

    # second object, based on the first one
    obj2 = object_store.new("b")
    obj2.init(tree)

    p = Path(obj2, "B")
    p.touch()

    obj2.finalize()  # put the object into READ mode
    assert obj2.mode == objectstore.Object.Mode.READ

    # commit always makes a copy, if space
    object_store.commit(tree, "b")

    assert object_store.contains("b")
    assert store_path(object_store, "b", "A")
    assert store_path(object_store, "b", "B")

    assert len(os.listdir(object_store.objects)) == 2

    # object should exist and should be in read mode
    tree = object_store.get("b")
    assert tree is not None
    assert tree.mode == objectstore.Object.Mode.READ


def test_cleanup(tmpdir):
    with objectstore.ObjectStore(tmpdir) as object_store:
        object_store.maximum_size = 1024*1024*1024

        stage = os.path.join(object_store, "stage")
        tree = object_store.new("a")
        assert len(os.listdir(stage)) == 1
        p = Path(tree, "A")
        p.touch()

    # there should be no temporary Objects dirs anymore
    with objectstore.ObjectStore(tmpdir) as object_store:
        assert object_store.get("A") is None


def test_metadata(tmpdir):

    # test metadata object directly first
    with tempfile.TemporaryDirectory() as tmp:
        md = objectstore.Object.Metadata(tmp)

        assert os.fspath(md) == tmp

        with pytest.raises(KeyError):
            with md.read("a"):
                pass

        # do not write anything to the file, it should not get stored
        with md.write("a"):
            pass

        assert len(list(os.scandir(tmp))) == 0

        # also we should not write anything if an exception was raised
        with pytest.raises(AssertionError):
            with md.write("a") as f:
                f.write("{}")
                raise AssertionError

        with md.write("a") as f:
            f.write("{}")

        assert len(list(os.scandir(tmp))) == 1

        with md.read("a") as f:
            assert f.read() == "{}"

    data = {
        "boolean": True,
        "number": 42,
        "string": "yes, please"
    }

    extra = {
        "extra": "data"
    }

    with tempfile.TemporaryDirectory() as tmp:
        md = objectstore.Object.Metadata(tmp)

        d = md.get("a")
        assert d is None

        md.set("a", None)
        with pytest.raises(KeyError):
            with md.read("a"):
                pass

        md.set("a", data)
        assert md.get("a") == data

    # use tmpdir fixture from here on
    with objectstore.ObjectStore(tmpdir) as store:
        store.maximum_size = 1024*1024*1024
        obj = store.new("a")
        p = Path(obj, "A")
        p.touch()

        obj.meta.set("md", data)
        assert obj.meta.get("md") == data

        store.commit(obj, "x")
        obj.meta.set("extra", extra)
        assert obj.meta.get("extra") == extra

        store.commit(obj, "a")

    with objectstore.ObjectStore(tmpdir) as store:
        obj = store.get("a")

        assert obj.meta.get("md") == data
        assert obj.meta.get("extra") == extra

        ext = store.get("x")

        assert ext.meta.get("md") == data
        assert ext.meta.get("extra") is None


@pytest.mark.skipif(not test.TestBase.can_bind_mount(), reason="Need root for bind mount")
def test_host_tree(tmpdir):
    with objectstore.ObjectStore(tmpdir) as store:
        host = store.host_tree

        assert host.tree
        assert os.fspath(host)

        # check we actually cannot write to the path
        p = Path(host.tree, "osbuild-test-file")
        with pytest.raises(OSError):
            p.touch()
            print("FOO")

    # We cannot access the tree property after cleanup
    with pytest.raises(AssertionError):
        _ = host.tree


def test_source_epoch(object_store):
    tree = object_store.new("a")
    tree.source_epoch = 946688461

    t = Path(tree)

    a = Path(tree, "A")
    a.touch()

    d = Path(tree, "d")
    d.mkdir()

    l = Path(tree, "l")
    l.symlink_to(d, target_is_directory=True)

    for i in (a, d, l, t):
        si = os.stat(i, follow_symlinks=False)
        before = int(si.st_mtime)
        assert before != tree.source_epoch

    # FINALIZING
    tree.finalize()

    for i in (a, d, l, t):
        si = os.stat(i, follow_symlinks=False)
        after = int(si.st_mtime)

        assert after != before, f"mtime not changed for {i}"
        assert after == tree.source_epoch

    baum = object_store.new("b")
    baum.init(tree)

    assert baum.created == tree.created

    b = Path(tree, "B")
    b.touch()

    si = os.stat(b, follow_symlinks=False)
    before = int(si.st_mtime)

    assert before != tree.source_epoch

    # FINALIZING
    baum.finalize()
    si = os.stat(a, follow_symlinks=False)
    after = int(si.st_mtime)

    assert after != before
    assert after == tree.source_epoch

# pylint: disable=too-many-statements


@pytest.mark.skipif(not test.TestBase.can_bind_mount(), reason="Need root for bind mount")
def test_store_server(tmpdir):
    with contextlib.ExitStack() as stack:

        store = objectstore.ObjectStore(tmpdir)
        stack.enter_context(store)

        tmp = tempfile.TemporaryDirectory()
        tmp = stack.enter_context(tmp)

        server = objectstore.StoreServer(store)
        stack.enter_context(server)

        client = objectstore.StoreClient(server.socket_address)

        have = client.source("org.osbuild.files")
        want = os.path.join(tmpdir, "sources")
        assert have.startswith(want)

        tmp = client.mkdtemp(suffix="suffix", prefix="prefix")
        assert tmp.startswith(store.tmp)
        name = os.path.basename(tmp)
        assert name.startswith("prefix")
        assert name.endswith("suffix")

        obj = store.new("42")
        p = Path(obj, "file.txt")
        p.write_text("osbuild")

        p = Path(obj, "directory")
        p.mkdir()
        obj.finalize()

        mountpoint = Path(tmpdir, "mountpoint")
        mountpoint.mkdir()

        assert store.contains("42")
        path = client.read_tree_at("42", mountpoint)
        assert Path(path) == mountpoint
        filepath = Path(mountpoint, "file.txt")
        assert filepath.exists()
        txt = filepath.read_text(encoding="utf8")
        assert txt == "osbuild"

        # check we can mount subtrees via `read_tree_at`

        filemount = Path(tmpdir, "file")
        filemount.touch()

        path = client.read_tree_at("42", filemount, "/file.txt")
        filepath = Path(path)
        assert filepath.is_file()
        txt = filepath.read_text(encoding="utf8")
        assert txt == "osbuild"

        dirmount = Path(tmpdir, "dir")
        dirmount.mkdir()

        path = client.read_tree_at("42", dirmount, "/directory")
        dirpath = Path(path)
        assert dirpath.is_dir()

        # check proper exceptions are raised for non existent
        # mount points and sub-trees

        with pytest.raises(RuntimeError):
            nonexistent = os.path.join(tmpdir, "nonexistent")
            _ = client.read_tree_at("42", nonexistent)

        with pytest.raises(RuntimeError):
            _ = client.read_tree_at("42", tmpdir, "/nonexistent")
